#include <errno.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include "rapidjson/document.h"
#include "NcClosers.h"
#include "Configuration.h"
#include "Constants.h"
#include "Messages.h"
#include "NcPrintErr.h"
#include "NcSocket.h"
#include "NcUart.h"

#include "SendTask.h"
#include "SendEssentials.h"

using namespace rapidjson;

bool ConfigureUart(tSendConfig * c)
{
	return UartConfigure(g_send.fd, c->getSpeed(), UART_READ_TIMEOUT_MS);
}

bool OpenSocket(tSendConfig * c)
{

	return SocketOpenServer(c->getSendPort(), &g_send.socketFd, &g_send.address);
}

bool OpenUart(tSendConfig * c)
{
	return UartOpen(c->getUart(), &g_send.fd);
}

bool SendTaskInit(void * p)
{
	Configuration * c = (Configuration *)p;
	SendTaskInitEssentials(c);
	g_send.config = c;
	if (!OpenUart(c) || !OpenSocket(c))
		return false;
	g_send.state = Started;
	return true;
}

bool SendTaskShutdown()
{
	g_send.state = Stopped;
	return true;
}

void SendTaskCloseUart()
{
	FdClose(&g_send.fd);
}

void SendTaskCloseSockets()
{
	FdClose(&g_send.socketFd);
}

void SendTaskClose()
{
	SendTaskCloseUart();
	SendTaskCloseSockets();
}

bool ReadClientMsg()
{
	int bufSize = SOCKET_INITIAL_BUFFER;
	uint8_t * p = (uint8_t*)malloc(bufSize);
	if (p==NULL)
	{
		puts("ERROR: out of memory, could not allocate socket buffer");
		return false;
	}
	int readSoFar = 0;
	int n;
	int c;
	int i;
	int TimeOut=50;
	
	c=0;
	while (true)
	{
		n = read(g_send.clientFd, p+readSoFar, bufSize-readSoFar);
		//printf("%d\n\r",n);
		if ((n==0)||(--TimeOut==0))
		{
			PrintErr("Failed to read from socket. no data");
			puts(g_error);
			free((void*)p);
			return false;
		
		}
		if (n<0)
		{
			switch(errno)
			{
			case EAGAIN:
			//case EWOULDBLOCK:
				break;
			default:
				PrintErr("Failed to read from socket. errno=%s", PrintErrNo(errno));
				puts(g_error);
				free((void*)p);
				return false;
			}
			usleep(100 * 1000);
			n=0;
		}
			
		for(i=0;i<n;i++)
		{
			switch( (p+readSoFar)[i])
			{
				case '{':
					c++;
					break;
				case '}':
					c--;
					break;
				default:
					break;
			}
		}
		
		readSoFar += n;
		if ((readSoFar!=0)&&(c==0))
			break;
		

		/*
		int n = read(g_send.clientFd, p+readSoFar, bufSize-readSoFar);
		bool finished = (n==0) || (n<0 && errno==EAGAIN) || (n<0 && errno==EWOULDBLOCK);
		if (finished)
			break;
		if (n<0 && !finished)
		{
			PrintErr("Failed to read from socket. errno=%s", PrintErrNo(errno));
			puts(g_error);
			free((void*)p);
			return false;
		}
		else
		{
			readSoFar += n;
			if (bufSize-readSoFar < 64)
			{
				p = (uint8_t *)realloc(p, bufSize + SOCKET_INITIAL_BUFFER);
				if (p==NULL)
				{
					puts("ERROR: out of memory, could not reallocate socket buffer");
					return false;
				}
			}
		}
		*/
	}
	g_send.clientRaw = p;
	g_send.clientRawLen = readSoFar;
	return true;
}

void CleanUpClient()
{
	FdClose(&g_send.clientFd);
	Free((void**)&g_send.clientRaw);
	Free((void**)&g_send.clientMsg);
	// Free((void**)&g_send.ncSendAckParams.payload);  // todo: ola
}

void WriteStatusToClient(const char * content)
{
	ReplyMsg rm =
		{
			content==NULL || strlen(content)==0 ? "ok" : "error",
			content
		};
	const char * ser = g_send.config->getSerializer()->Serialize(&rm);
	printf("-->%s\n", ser);
	const char * del = g_send.config->getDelimiterFormatted();
	int delLen = strlen(del);
	int len = strlen(ser);
	int finalLen = len + delLen;
	char * final = (char *)malloc(finalLen);
	memcpy(final, ser, len);
	memcpy(final+len, del, delLen);
	// dummy, the posix write is declared with __WUR
	ssize_t dummy = send(g_send.clientFd, final, finalLen, MSG_NOSIGNAL);
	(void) dummy;

	free((void*)ser);
	free((void*)final);
}

bool WaitForMessageToBeWritten()
{
	const int timeOutSecs = g_send.config->getCtsTimeoutSecs();
	struct timespec tvend;
	struct timespec tv;
	clock_gettime(CLOCK_MONOTONIC_RAW, &tvend);
	tvend.tv_sec += timeOutSecs;
	while(!g_send.msgWasWritten)
	{
		usleep(SOCKET_LOOP_PAUSE_MS);
		clock_gettime(CLOCK_MONOTONIC_RAW, &tv);
		if (tv.tv_sec>tvend.tv_sec)
			return false;
	}
	return true;
}

void SendTaskHandleClient(int fd)
{
	g_send.clientFd = fd;
	if (!ReadClientMsg())
		return;
	PrintMsg();

	const char * content = ParseMsg();
	if (content==NULL)
	{
		NcApiErrorCodes ncRet = SendMessageToNcApi();
		if (ncRet != NCAPI_OK)
		{
			PrintErr("Staging data failed: %d", ncRet);
			content = g_error;
		}
		if (!WaitForMessageToBeWritten())
		{
			NcApiCancelEnqueuedMessage(0);
			const int timeOutSecs = g_send.config->getCtsTimeoutSecs();
			PrintErr("Message was not written to UART within %d seconds", timeOutSecs);
			content = g_error;
		}
	}
	WriteStatusToClient(content);
	CleanUpClient();
}

void SendTaskAcceptSockets()
{
	int clientFd;
	while (SocketAcceptNonBlock(g_send.socketFd, &clientFd))
		SendTaskHandleClient(clientFd);
}

void SendTaskSleep()
{
	usleep(SOCKET_LOOP_PAUSE_MS * 1000);
}

void SendTaskLoop()
{
	SendTaskAcceptSockets();
	SendTaskSleep();
}

void* SendTaskRun(void * p)
{
	tSendConfig * c = (tSendConfig *)p;
    if (!ConfigureUart(c))
    	return NULL;
	g_send.state = Started;
	while(true)
	{
		switch (g_send.state)
		{
		case Stopped:
			SendTaskClose();
			return NULL;
		case Started:
			SendTaskLoop();
			break;
		case ReLoadConfig:
			SendTaskClose();
			g_send.state = SendTaskInit(g_send.config)
					? Started
					: Stopped;
			break;
		}
	}
	return NULL;
}

NcApiErrorCodes NcApiSupportTxData(uint8_t n, uint8_t * finalMsg, uint8_t finalMsgLength)
{
	// dummy, the posix write is declared with __WUR
	ssize_t dummy = write(g_send.fd, finalMsg, finalMsgLength);
	(void) dummy;
	return NCAPI_OK;
}

void NcApiSupportMessageWritten(uint8_t n, void * callbackToken, uint8_t * finalMsg, uint8_t finalMsgLength)
{
	g_send.msgWasWritten = true;
}
