#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include "rapidjson/document.h"
#include "rapidjson/stringbuffer.h"
#include "rapidjson/writer.h"
#include "Serializer.h"
#if !_WIN32 && !_WIN64 // Exclude this when target is a Windows platform
#include <langinfo.h>
#endif
#include <time.h>
#include <string.h>

using namespace rapidjson;
using namespace std;

static const uint8_t Nibbles[16] = {'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};


// Define actual serializer functions for the JSON format


typedef class tSerializerJson : public ISerializer
{
	const char * ToHexString(uint8_t * ba, uint8_t len)
	{
		char* dst = (char*)malloc(2*len+1);
		int i;
		for (i=0; i<len; i++)
		{
			uint8_t b = ba[i];
			dst[i*2+0] = Nibbles[ b >> 4 ];
			dst[i*2+1] = Nibbles[ b & 0xf ];
		}
		dst[ len << 1 ] = 0;
		return dst;
	}

	void WritePayloadRaw(Writer<StringBuffer>& w, uint8_t * ba, uint8_t len)
	{
		w.String("payload");
		w.StartArray();
		for (unsigned i = 0; i < len; i++)
		    w.Uint(ba[i]);
		w.EndArray();
	}


	void WriteMsgType(Writer<StringBuffer>& w, NcApiMessageType mt)
	{

		w.String("objectType");
		switch(mt)
		{
		case HostAckEnum:
			w.String("ack");
			break;
		case HostNAckEnum:
			w.String("nack");
			break;
		case HostDataEnum:
		case HostDataHapaEnum:
		case HostUappDataEnum:
		case HostUappDataHapaEnum:
			w.String("receivedPayload");
			break;
		case NodeInfoReplyEnum:
			w.String("nodeInfoReply");
			break;
		case NeighborListReplyEnum:
			w.String("neighborListReply");
			break;
		case NetCmdReplyEnum:
			w.String("networkCommandReply");
			break;
		case WesStatusEnum:
			w.String("wesStatus");
			break;
		case WesSetupRequestEnum:
			w.String("wesSetupRequest");
			break;
		default:
			w.String("unknown");
			break;
		}
	}

	void WritePayloadType(Writer<StringBuffer>& w, bool acknowledged)
	{
		w.String("payloadType");
		if (acknowledged)
			w.String("acknowledged");
		else
			w.String("unacknowledged");
	}

	void WriteSeqNo(Writer<StringBuffer>& w, uint16_t v)
	{
		w.String("sequenceNo");
		w.Uint(v);
	}

	void WriteNodeType(Writer<StringBuffer>& w, uint8_t v)
	{
		w.String("nodeType");
		w.Uint(v);
	}

	void WriteGwTimestamp(Writer<StringBuffer>& w)
	{
		char datestring[32];
		char datestringEx[32];
		long ms = 0; // Milliseconds

#if !_WIN32 && !_WIN64 // Exclude this when target is a Windows platform
		struct timespec ts;

		timespec_get(&ts, TIME_UTC);
		ms = (ts.tv_nsec / 1.0e6); // Convert nanoseconds to milliseconds
	
		strftime(datestring, sizeof datestring, "%FT%T", gmtime(&ts.tv_sec));
#else
		// Just create a fixed dummy string
		sprintf(datestring, "2017-01-01T12:00:00");
#endif

		sprintf(datestringEx,"%s.%03ldZ",datestring,ms % 1000);
		w.String("gwTimestamp");
		w.String(datestringEx);
	}

	void WriteOriginId(Writer<StringBuffer>& w, uint16_t v)
	{
		w.String("nodeId");
		w.Uint(v);
	}

	void WriteUid(Writer<StringBuffer>& w, uint8_t v[5])
	{
		const char * p = ToHexString(v, 5);
		w.String("uidHex");
		w.String(p);
		free((void*)p);
	}

	void WritePort(Writer<StringBuffer>& w, uint8_t v)
	{
		w.String("port");
		w.Uint(v);
	}

	void WritePackageAge(Writer<StringBuffer>& w, uint16_t v)
	{
		w.String("packageAgeMicroSecs");
		w.Uint64((uint64_t(v)) * 125000UL);
	}

	void WritePackageAgeHapa(Writer<StringBuffer>& w, uint32_t v)
	{
		w.String("packageAgeMicroSecs");
		w.Uint64((uint64_t(v)) * 1000000UL / 524288UL);
	}

	void WriteStatus(Writer<StringBuffer>& w, const char * s)
	{
		w.String("status");
		w.String(s);
	}

	void WriteContent(Writer<StringBuffer>& w, const char * s)
	{
		w.String("content");
		w.String(s);
	}

	void WriteWesStatus(Writer<StringBuffer>& w, NcApiWesStatusValues v)
	{
		w.String("status");
		w.Uint(v);
	}

	void WriteAppFuncType(Writer<StringBuffer>& w, uint8_t v)
	{
		w.String("appFunctionType");
		w.Uint(v);
	}

	void WriteRSSI(Writer<StringBuffer>& w, uint8_t v)
	{
		w.String("RSSI");
		w.Uint(v);
	}

	void WriteCommand(Writer<StringBuffer>& w, uint8_t v)
	{
		w.String("command");
		w.Uint(v);
	}

	const char * ParseJson(const char * json, int jsonLen, Document * d)
	{
#if !_WIN32 && !_WIN64 // Exclude this when target is a Windows platform
		const char * m = strndup(json, jsonLen);
#else
		const char * m = json;
#endif

		if (*m!='{')
		{
#if !_WIN32 && !_WIN64 // Exclude this when target is a Windows platform
			free((void*)m);
#endif
			return "Message is not JSON-formatted (#1)";
		}
		if (d->Parse(m).HasParseError())
		{
#if !_WIN32 && !_WIN64 // Exclude this when target is a Windows platform
			free((void*)m);
#endif
			return "Message is not JSON-formatted (#2)";
		}
#if !_WIN32 && !_WIN64 // Exclude this when target is a Windows platform
		free((void*)m);
#endif
		return NULL;
	}

	const char * RaiseMissingMember(const char * member)
	{
		char * p = (char*)malloc(1024);
		sprintf(p, "Missing member: %s", member);
		return p;
	}

	const char * ReadUInt8(Document& d, const char * member, uint8_t * p, bool raiseOnError = true)
	{
		if (!d.HasMember(member))
		{
			if (raiseOnError)
				return RaiseMissingMember(member);
			return NULL;
		}
		*p = d[member].GetUint();
		return NULL;
	}

	const char * ReadUInt16(Document& d, const char * member, uint16_t * p, bool raiseOnError = true)
	{
		if (!d.HasMember(member))
		{
			if (raiseOnError)
				return RaiseMissingMember(member);
			return NULL;
		}
		*p = d[member].GetUint();
		return NULL;
	}

	const char * ReadUInt32(Document& d, const char * member, int * p, bool raiseOnError = true)
	{
		if (!d.HasMember(member))
		{
			if (raiseOnError)
				return RaiseMissingMember(member);
			return NULL;
		}
		*p = d[member].GetUint();
		return NULL;
	}

	uint8_t HexCharToNibble(char c)
	{
		if (c<='9') return c - '0';
		if (c<='F') return c - 'A' + 10;
		return c - 'a' + 10;
	}

	const char * ReadHexString(Document& d, const char * member, uint8_t * payload, uint8_t len, bool raiseOnError = true)
	{
		if (!d.HasMember(member))
		{
			if (raiseOnError)
				return RaiseMissingMember(member);
			return NULL;
		}
		const char * hex = d[member].GetString();
		int hexlen = strlen(hex);
		if ((hexlen >> 1) != len)
			return strdup("HexString length mismatch");
		int i;
		for (i=0; i<len; i++)
			payload[i] = (HexCharToNibble(hex[2*i+0]) << 4)
					   | (HexCharToNibble(hex[2*i+1]));
		return NULL;
	}

	const char * ReadString(Document& d, const char * member)
	{
		if (!d.HasMember(member))
		{
			return NULL;
		}
		return d[member].GetString();
	}

	const char * ReadIntArray(Document& d, const char * member, uint8_t ** payload, uint8_t * len, bool raiseOnError = true)
	{
		if (!d.HasMember(member))
		{
			if (raiseOnError)
				return RaiseMissingMember(member);
			return NULL;
		}
		if (!d[member].IsArray())
		{
			if (raiseOnError)
				return RaiseMissingMember(member);
			return NULL;
		}
		Value& p = d[member];
		*len = p.Size();
		if (*len == 0)
		{
			*payload = NULL;
			return NULL;
		}
		*payload = (uint8_t*)malloc(*len);
		for (SizeType i = 0; i < p.Size(); i++) // Uses SizeType instead of size_t
		        (*payload)[i]=p[i].GetInt();

		return NULL;
	}

public:
	virtual ~tSerializerJson() {}
	virtual const char * SerializeAck(tNcApiHostAckNack * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, HostAckEnum);
		WriteGwTimestamp(w);
		WriteOriginId(w, p->originId);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * SerializeNAck(tNcApiHostAckNack * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, HostNAckEnum);
		WriteGwTimestamp(w);
		WriteOriginId(w, p->originId);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(tNcApiHostData * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, HostDataEnum);
		WriteGwTimestamp(w);
		WritePayloadType(w, true);
		WriteOriginId(w, p->originId);
		WritePort(w, p->port);
		WritePackageAge(w, p->packageAge);
		w.String("packageAgeType");
		w.String("normal");		
    	WritePayloadRaw(w, p->payload, p->payloadLength);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(tNcApiHostUappData * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, HostUappDataEnum);
		WriteGwTimestamp(w);
		WritePayloadType(w, false);
		WriteSeqNo(w, p->appSeqNo);
		WriteOriginId(w, p->originId);
		WritePort(w, p->port);
		WritePackageAge(w, p->packageAge);
		w.String("packageAgeType");
		w.String("normal");		
    	WritePayloadRaw(w, p->payload, p->payloadLength);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * SerializeRaw(uint8_t * Buffer,uint8_t payloadLength)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, (NcApiMessageType)Buffer[0]);
		WriteGwTimestamp(w);
    	WritePayloadRaw(w, &Buffer[0], payloadLength);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(tNcApiHostDataHapa * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, HostDataHapaEnum);
		WriteGwTimestamp(w);
		WritePayloadType(w, true);
		WriteOriginId(w, p->originId);
		WritePort(w, p->port);
		WritePackageAgeHapa(w, p->packageAge);
		w.String("packageAgeType");
		w.String("hapa");		
		WritePayloadRaw(w, p->payload, p->payloadLength);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(tNcApiHostUappDataHapa * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, HostUappDataHapaEnum);
		WriteGwTimestamp(w);
		WritePayloadType(w, false);
		WriteSeqNo(w, p->appSeqNo);
		WriteOriginId(w, p->originId);
		WritePort(w, p->port);
		WritePackageAgeHapa(w, p->packageAge);
		w.String("packageAgeType");
		w.String("hapa");		
		WritePayloadRaw(w, p->payload, p->payloadLength);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(tNcApiNodeInfoReply * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, NodeInfoReplyEnum);
		WriteGwTimestamp(w);
		WriteOriginId(w, p->nodeId);
		WriteUid(w, p->uid);
		WriteNodeType(w, p->Type);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(tNcApiNeighborListReply * p)
	{
		int i;
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, NeighborListReplyEnum);
		WriteGwTimestamp(w);

		w.String("neighbors");
		w.StartArray();
		for (i = 0; i < p->NeighborsCount; i++)
		{
			w.StartObject();
			WriteOriginId(w, p->Neighbor[i].nodeId);
			WriteRSSI(w, p->Neighbor[i].RSSI);
			w.EndObject();
		}
		w.EndArray();

		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(tNcApiNetCmdReply * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, NetCmdReplyEnum);
		WriteGwTimestamp(w);
		WriteOriginId(w, p->originId);
		WriteCommand(w, p->cmd);
    	WritePayloadRaw(w, p->payload, p->payloadLength);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(tNcApiWesSetupRequest * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, WesSetupRequestEnum);
		WriteGwTimestamp(w);
		WriteUid(w, p->uid);
		WriteAppFuncType(w, p->appFuncType);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(tNcApiWesStatus * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, WesStatusEnum);
		WriteGwTimestamp(w);
		WriteWesStatus(w, (NcApiWesStatusValues)p->Status);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(ReplyMsg * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		
		w.String("objectType");
		w.String("replyMsg");
				
		WriteGwTimestamp(w);
		
		
		WriteStatus(w, p->status);
		if (p->content==NULL)
			WriteContent(w, "");
		else
			WriteContent(w, p->content);
		
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * DeSerialize(const char * m, int len, tNcApiSendUnackMessage * msg)
	{
		Document d;
		const char * ret = ParseJson(m, len, &d);
		const char * str; 
		const char * str_2; 
	
		if (ret==NULL)
		{
			str=ReadString(d,"objectType");
			str_2=ReadString(d,"payloadType");
#ifdef GATEWAY_DEBUG_INFORMATION
			printf("      DeSerialize/Send(unacknowledged), str %s, str_2 %s \r\n", str, str_2);
#endif
			if (str!=NULL && str_2!=0)
			{
				if (strcmp(str,"sendPayload")==0 &&
					strcmp(str_2,"unacknowledged")==0)
				{
					ret = ReadIntArray(d, "payload", &msg->payload, &msg->payloadLength);
					if (ret==NULL) ret = ReadUInt16(d, "nodeId", &msg->destNodeId);
					if (ret==NULL) ret = ReadUInt8(d, "port", &msg->destPort);
					if (ret==NULL) ret = ReadUInt16(d, "sequenceNo", &msg->appSeqNo);
				}
				else
				{
					ret="";
				}
			}
			else
			{
				ret="";
			}
		}
	
		return ret;
	}

	virtual const char * DeSerialize(const char * m, int len, tNcApiSendAckMessage * msg)
	{
		Document d;
		const char * ret = ParseJson(m, len, &d);
		const char * str; 
		const char * str_2; 
	
		if (ret==NULL)
		{
			str=ReadString(d,"objectType");
			str_2=ReadString(d,"payloadType");
#ifdef GATEWAY_DEBUG_INFORMATION
			printf("      DeSerialize/Send(acknowledged), str %s, str_2 %s \r\n", str, str_2);
#endif
			if (str!=NULL && str_2!=0)
			{
				if (strcmp(str,"sendPayload")==0 &&
					strcmp(str_2,"acknowledged")==0)
				{
					ret = ReadIntArray(d, "payload", &msg->payload, &msg->payloadLength);
					if (ret==NULL) ret = ReadUInt16(d, "nodeId", &msg->destNodeId);
					if (ret==NULL) ret = ReadUInt8(d, "port", &msg->destPort);
				}
				else
				{
					ret="";
				}
			}
			else
			{
				ret="";
			}
		}
	
		return ret;
	}

	virtual const char * DeSerialize(const char * m, int len, tNcApiNodeInfoRequestMessage * msg)
	{
		Document d;
		const char * ret = ParseJson(m, len, &d);
		const char * str; 

		if (ret==NULL)
		{
			str=ReadString(d,"objectType");
#ifdef GATEWAY_DEBUG_INFORMATION
			printf("      DeSerialize/nodeInfoRequest, str %s \r\n", str);
#endif
			if (str!=NULL)
			{
				if (strcmp(str,"nodeInfoRequest")==0)
				{
					;
				}
				else
				{
					ret="";
				}
			}
			else
			{
				ret="";
			}
		}

		return ret;
	}

	virtual const char * DeSerialize(const char * m, int len, tNcApiNeighborListRequestMessage * msg)
	{
		Document d;
		const char * ret = ParseJson(m, len, &d);
		const char * str; 

		if (ret==NULL)
		{
			str=ReadString(d,"objectType");
#ifdef GATEWAY_DEBUG_INFORMATION
			printf("      DeSerialize/neighborListRequest, str %s \r\n", str);
#endif
			if (str!=NULL)
			{
				if (strcmp(str,"neighborListRequest")==0)
				{
					;
				}
				else
				{
					ret="";
				}
			}
			else
			{
				ret="";
			}
		}

		return ret;
	}

	virtual const char * DeSerialize(const char * m, int len, tNcApiNetCmdMessage * msg)
	{
		Document d;
		const char * ret = ParseJson(m, len, &d);
		const char * str; 

		if (ret==NULL)
		{
			str=ReadString(d,"objectType");
#ifdef GATEWAY_DEBUG_INFORMATION
			printf("      DeSerialize/networkCommand, str %s \r\n", str);
#endif
			if (str!=NULL)
			{
				if (strcmp(str,"networkCommand")==0)
				{
					ret = ReadIntArray(d, "payload", &msg->payload, &msg->payloadLength);
					if (ret==NULL) ret = ReadUInt16(d, "nodeId", &msg->destNodeId);
					if (ret==NULL) ret = ReadUInt8(d, "cmd", &msg->cmd);
				}
				else
				{
					ret="";
				}
			}
			else
			{
				ret="";
			}
		}

		return ret;
	}

	virtual const char * DeSerialize(const char * m, int len, tNcApiWesCmdMessage * msg)
	{
		Document d;
		const char * ret = ParseJson(m, len, &d);
		uint8_t byteValue = 0;
		const char * str; 

		if (ret==NULL)
		{
			str=ReadString(d,"objectType");
#ifdef GATEWAY_DEBUG_INFORMATION
			printf("      DeSerialize/wesCmd, str %s \r\n", str);
#endif
			if (str!=NULL)
			{
				if (strcmp(str,"wesCmd")==0)
				{
					ret = ReadUInt8(d, "cmd", &byteValue);
					if (ret==NULL) msg->cmd = (NcApiWesCmdValues)byteValue;
				}
				else
				{
					ret="";
				}
			}
			else
			{
				ret="";
			}
		}

		return ret;
	}

	virtual const char * DeSerialize(const char * m, int len, tNcApiWesResponseMessage * msg)
	{
		Document d;
		const char * ret = ParseJson(m, len, &d);
		const char * str; 
		
		if (ret==NULL)
		{
			str=ReadString(d,"objectType");
#ifdef GATEWAY_DEBUG_INFORMATION
			printf("      DeSerialize/wesResponse, str %s \r\n", str);
#endif
			if (str!=NULL)
			{
				if (strcmp(str,"wesResponse")==0)
				{
					ret = ReadUInt16(d, "nodeId", &msg->nodeId);
					if (ret==NULL) ret = ReadHexString(d, "uidHex", msg->uid, (uint8_t)sizeof(msg->uid));
					if (ret==NULL) ret = ReadHexString(d, "appSettings", msg->appSettings, (uint8_t)sizeof(msg->appSettings));
				}
				else
				{
					ret="";
				}
			}
			else
			{
				ret="";
			}
		}

		return ret;
	}

	virtual const char * DeSerialize(const char * m, int len, tNcApiAltCmdMessage * msg)
	{
		Document d;
		const char * ret = ParseJson(m, len, &d);
		uint8_t byteValue = 0;
		const char * str; 

		if (ret==NULL)
		{
			str=ReadString(d,"objectType");
#ifdef GATEWAY_DEBUG_INFORMATION
			printf("      DeSerialize/altCmd, str %s \r\n", str);
#endif
			if (str!=NULL)
			{
				if (strcmp(str,"altCmd")==0)
				{
					ret = ReadUInt8(d, "cmd", &byteValue);
					if (ret==NULL) msg->cmd = (NcApiAltCmdValues)byteValue;
				}
				else
				{
					ret="";
				}
			}
			else
			{
				ret="";
			}
		}

		return ret;
	}


} SerializerJson;
